"""Модуль определяющий стандартные реализованные события."""
from pywood.events import BaseEvent
from pywood.updates.commands import is_command, is_start_command
from pywood.updates.numbers import is_int_number
from pywood.updates.texts import is_msg_w_text_to_bot

__all__ = [
    'IncomingText',
    'IncomingCommand',
    'IncomingIntNumber',
    'IncomingUpdate',
    'IncomingStartCommand',
]


class IncomingText(BaseEvent):
    """Входящий текст."""

    @staticmethod
    def happened(update) -> bool:
        return is_msg_w_text_to_bot(update)


class IncomingUpdate(BaseEvent):
    """Любое входящее обновление."""

    @staticmethod
    def happened(update) -> bool:
        return True


class IncomingCommand(BaseEvent):
    """Входящая команда."""

    @staticmethod
    def happened(update) -> bool:
        return is_command(update)

class IncomingStartCommand(BaseEvent):
    """Входящая команда `start`."""

    @staticmethod
    def happened(update) -> bool:
        return is_start_command(update)


class IncomingIntNumber(BaseEvent):
    """Число."""

    @staticmethod
    def happened(update) -> bool:
        return is_int_number(update)



import unittest

from pywood.updates.numbers import is_int_number


class TestIsIntNumber(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        class Update:
            def __init__(self, message, value1, value2):
                self.message = message
                self.value1 = value1
                self.value2 = value2

        class Message:
            def __init__(self, text, value1, value2):
                self.text = text
                self.value1 = value1
                self.value2 = value2

        cls.Update = Update
        cls.Message = Message

    def test_is_int_number_assert_true(self):
        message_with_text_with_int_1 = self.Message('123', None, 23432)
        message_with_text_with_int_2 = self.Message('234', None, None)
        message_with_text_with_int_3 = self.Message('-3456', 'van', 'rossum')
        message_with_text_with_int_4 = self.Message('0', 'van', 'rossum')

        update_with_msg_with_text_with_int_1 = self.Update(message_with_text_with_int_1,
                                                           'typing',
                                                           123)

        update_with_msg_with_text_with_int_2 = self.Update(message_with_text_with_int_2,
                                                           None,
                                                           123)

        update_with_msg_with_text_with_int_3 = self.Update(message_with_text_with_int_3,
                                                           None,
                                                           'van')

        update_with_msg_with_text_with_int_4 = self.Update(message_with_text_with_int_4,
                                                           100000,
                                                           -345678)

        self.assertTrue(is_int_number(update_with_msg_with_text_with_int_1))
        self.assertTrue(is_int_number(update_with_msg_with_text_with_int_2))
        self.assertTrue(is_int_number(update_with_msg_with_text_with_int_3))
        self.assertTrue(is_int_number(update_with_msg_with_text_with_int_4))

    def test_is_int_number_assert_false(self):
        message_with_text_wo_int_1 = self.Message('123.123', None, 23432)
        message_with_text_wo_int_2 = self.Message('234234324.3423', None, None)
        message_with_text_wo_int_3 = self.Message('Elon Musk', 'van', 'rossum')
        message_with_text_wo_int_4 = self.Message('-123123.00001', 'van', 'rossum')

        message_wo_text_1 = self.Message(None, None, None)
        message_wo_text_2 = self.Message(None, 'germany', None)
        message_wo_text_3 = self.Message(None, 'belarus', 'dzmitry')

        update_with_msg_wo_text1 = self.Update(message_wo_text_1,
                                               'rossum',
                                               123)

        update_with_msg_wo_text2 = self.Update(message_wo_text_2,
                                               None,
                                               123)

        update_with_msg_wo_text3 = self.Update(message_wo_text_3,
                                               None,
                                               'guido')

        update_with_msg_wo_int_1 = self.Update(message_with_text_wo_int_1,
                                               'Jeff',
                                               'guido')

        update_with_msg_wo_int_2 = self.Update(message_with_text_wo_int_2,
                                               'Bezos',
                                               'van')

        update_with_msg_wo_int_3 = self.Update(message_with_text_wo_int_3,
                                               'Me',
                                               'rossum')

        update_with_msg_wo_int_4 = self.Update(message_with_text_wo_int_3,
                                               'Dzmitry',
                                               'Maliuzhenets')

        update_wo_msg_1 = self.Update(None, None, None)
        update_wo_msg_2 = self.Update(None, "elon", 123)
        update_wo_msg_3 = self.Update(None, "elon", "musk")

        self.assertFalse(is_int_number(update_wo_msg_1))
        self.assertFalse(is_int_number(update_wo_msg_3))
        self.assertFalse(is_int_number(update_wo_msg_2))
        self.assertFalse(is_int_number(update_with_msg_wo_text2))
        self.assertFalse(is_int_number(update_with_msg_wo_text3))
        self.assertFalse(is_int_number(update_with_msg_wo_int_1))
        self.assertFalse(is_int_number(update_with_msg_wo_int_2))
        self.assertFalse(is_int_number(update_with_msg_wo_int_3))
        self.assertFalse(is_int_number(update_with_msg_wo_int_4))



import random
import string
from typing import Optional, List, Callable

"""Модуль для генерации объектов Update для целей тестирования."""


class Message:
    def __init__(self, chat, text, entities=None):
        self.chat = chat
        self.text = text
        self.entities = entities


class Chat:
    def __init__(self, id):
        self.id = id


class Update:
    def __init__(self, message):
        self.message = message


def update_w_random_int_number(chat_id):
    chat = Chat(chat_id)
    random_int = random.randint(-1000000, 1000000)
    message = Message(chat, str(random_int))
    return Update(message=message)


def update_w_random_update_text(chat_id):
    def random_string(string_length=10):
        """Generate a random string of fixed length """
        letters = string.ascii_lowercase
        return ''.join(random.choice(letters) for i in range(string_length))

    chat = Chat(chat_id)
    message = Message(chat, random_string(50))
    return Update(message=message)


def update_w_random_command(chat_id):
    class MessageEntity:
        def __init__(self, type):
            self.type = type

    def random_string(string_length=20):
        """Generate a random string of fixed length """
        letters = string.ascii_lowercase
        return '/' + ''.join(
            random.choice(letters) for i in range(string_length))

    chat = Chat(chat_id)
    message = Message(chat, random_string(50),
                      entities=[MessageEntity(type='bot_command')])
    return Update(message=message)


def get_random_update(exclude_chat_ids: Optional[List[int]] = None,
                      funcs: Optional[List[Callable]] = None,
                      chat_id_range=(-999999999999, 999999999999)) -> Update:
    """Сгенерировать случайное Update.

    :param exclude_chat_ids: (optional) список chat_id, которые нельзя использовать для генерации
    :type exclude_chat_ids: List[int]

    :param funcs:  (optional) список функций, которые будут участвовать в генерации объекта Update; функции должны
    принимать аргументом chat_id и возвращать объект Update
    :type funcs: List[Callable]

    :param chat_id_range: (optional) диапазон генерируемых chat_id
    :type chat_id_range: Tuple[int, int]
    """

    if not funcs:
        funcs = [
            update_w_random_int_number,
            update_w_random_update_text,
            update_w_random_command,
        ]

    func = random.choice(funcs)
    while True:
        rand_chat_id = random.randint(chat_id_range[0],
                                      chat_id_range[1])
        if exclude_chat_ids:
            if rand_chat_id in exclude_chat_ids:
                continue
            else:
                return func(rand_chat_id)
        else:
            return func(rand_chat_id)



import unittest
from pywood.updates.generate import update_w_random_command
from pywood.updates.generate import update_w_random_int_number
from pywood.updates.generate import update_w_random_update_text
from pywood.updates.texts import is_msg_w_text_to_bot
from pywood.updates.commands import is_command
from pywood.updates.numbers import is_int_number
from pywood.updates.generate import get_random_update


class Test(unittest.TestCase):
    def test_update_w_random_int_number(self):
        self.assertTrue(is_int_number(update_w_random_int_number(123123123)))

    def test_update_w_random_update_text(self):
        self.assertTrue(is_msg_w_text_to_bot(update_w_random_update_text(123123123)))

    def test_update_w_random_command(self):
        self.assertTrue(is_command(update_w_random_command(123123123)))


class TestGetRandomUpdate(unittest.TestCase):

    def test_chat_ids(self):
        chat_ids_for_exclude = []
        chat_id_range = (-1000, 1000)
        for i in range(chat_id_range[0], chat_id_range[1], 4):
            chat_ids_for_exclude.append(i)
        for _ in range(1000):
            update = get_random_update(exclude_chat_ids=chat_ids_for_exclude,
                                       chat_id_range=chat_id_range)
            chat_id = update.message.chat.id
            self.assertNotIn(chat_id, chat_ids_for_exclude)





from .commands import is_command


def is_msg_w_text_to_bot(update) -> bool:
    if is_command(update):
        return False

    try:
        update.message.text
    except AttributeError:
        return False
    else:
        return update.message.text is not None


def is_channel_post_w_text(update) -> bool:
    if is_command(update):
        return False

    try:
        update.channel_post.text
    except AttributeError:
        return False
    else:
        return update.channel_post.text is not None



import re

int_regexp = re.compile(r"[-+]?\d+$")


def _is_int(s):
    return int_regexp.match(s) is not None


def is_int_number(update) -> bool:
    try:
        text = update.message.text.strip()
    except (AttributeError, ValueError, TypeError):
        return False
    else:
        return _is_int(text)



"""Класс определяющий основную функциональность."""
from __future__ import annotations

from typing import List, Type, Callable, TypeVar, Union, Tuple, Any

from pywood.errors import StateDoesNotExistError
from pywood.states import BaseState

Update = TypeVar('Update')


class Machine:
    """Класс определяющий работу с состояниями."""

    def __init__(self,
                 *,
                 current_state_getter: Callable,
                 states: List[Type[BaseState]]):
        """Проинициализировать класс определяющий работу с состояниями.

        :param current_state_getter: функция, принимающая Update и \
        возвращающая либо строку, которая является названием текущего \
        состояния, либо сам класс текущего состояния
        :type current_state_getter: функция

        :param states: список состояний чата
        :type states: список из классов унаследованных от State
        """
        self.current_state_getter = current_state_getter
        self.states = states

    def get_state_from_state_name(self, state_name: str) -> Type[BaseState]:
        """Получить класс состояния из названия класса состояния."""
        for state in self.states:
            if state.__name__ == state_name:
                return state
        raise StateDoesNotExistError

    def get_current_state(self, state) -> Tuple[Type[BaseState], Any]:
        """Получить текущее состояние"""
        if isinstance(state, str):
            return self.get_state_from_state_name(state), None
        elif isinstance(state, tuple):
            if isinstance(state[0], str):
                return self.get_state_from_state_name(state[0]), state[1]
            elif issubclass(state[0], BaseState):
                return state[0], state[1]
            else:
                raise StateDoesNotExistError

        elif issubclass(state, BaseState):
            return state, None
        else:
            raise StateDoesNotExistError(
                "`current_state_getter` должна возвращать"
                " строку, класс или кортеж вида (str, Any)")

    @staticmethod
    def add_attrs(obj, attrs):
        for key, value in attrs.items():
            setattr(obj, key, value)

    def process(self,
                update: Update,
                state_attrs=None,
                ):
        """Обработать объект типа Update."""
        raw_curr_state = self.current_state_getter(update)
        state_cls, state_data = self.get_current_state(raw_curr_state)
        state = state_cls()
        if state_attrs:
            Machine.add_attrs(state, state_attrs)
        state._traverse_handlers(update, state_data)



from typing import Optional, Dict, Any, Callable, List, Type

from pywood._machine import Machine
from pywood.states import BaseState


def handle_update(update,
                  states: List[Type[BaseState]],
                  current_state_getter: Callable,
                  state_attrs: Optional[Dict[str, Any]] = None,
                  ):
    """

    :param update: обновление
    :type update: объект, определяющий объект `Update <https://core.telegram.org/bots/api#update>`_

    :param states: список классов состояний
    :type states: список классов, наследующих от BaseState (List[Type[BaseState]])

    :param current_state_getter: callable, принимающее как аргумент `update`,
        и возвращающее

        - строку, содержащую название класса текущего состояния
        - сам класс текущего состояния
        - кортеж, содержащий первым своим элементом название класса текущего
          состояния и вторым элементом дополнительную информацию, которая будет
          добавлена в аргумент `context` обработчиков

    :type current_state_getter: callable

    :param state_attrs: (optional) словарь, содержащий аттрибуты, которые будут
        добавлены к экземпляру класса состояния
    :type state_attrs: Dict[str, Any]

    :raises StateDoesNotExistError: выбрасывается, если current_state_getter
        возвращает не строку, класс состояния или кортеж

    :return:
    """
    machine = Machine(current_state_getter=current_state_getter,
                      states=states)
    machine.process(update=update, state_attrs=state_attrs)



"""Module for testing the `event` decorator."""
import unittest
from unittest.mock import MagicMock

from pywood import decorators
from pywood.context import Context
from pywood.decorators import events, no_events, event
from pywood.events import BaseEvent
from pywood._exceptions import EventHandled
from pywood.states import BaseState


class EventTrue1(BaseEvent):
    @staticmethod
    def happened(update) -> bool:
        return True

class EventTrue2(BaseEvent):
    @staticmethod
    def happened(update) -> bool:
        return True

class EventTrue3(BaseEvent):
    @staticmethod
    def happened(update) -> bool:
        return True


class EventFalse1(BaseEvent):
    @staticmethod
    def happened(update) -> bool:
        return False

class EventFalse2(BaseEvent):
    @staticmethod
    def happened(update) -> bool:
        return False

class EventFalse3(BaseEvent):
    @staticmethod
    def happened(update) -> bool:
        return False

class EventFalse4(BaseEvent):
    @staticmethod
    def happened(update) -> bool:
        return False


class TestNoEventsDecorator(unittest.TestCase):
    """Testing the `event` decorator."""

    def test_method_called_once_with(self):

        class State(BaseState):
            def handler1(self, update, event_cls, data):
                pass

            def handler2(self, update, event_cls, data):
                pass

            def handler3(self, update, data):
                pass

        handler1_mock = MagicMock(__name__='handler1')
        handler2_mock = MagicMock(__name__='handler2')
        handler3_mock = MagicMock(__name__='handler3')
        State.handler1 = events(EventFalse1, EventFalse2)(handler1_mock)
        State.handler2 = events(EventFalse3, EventFalse4)(handler2_mock)
        State.handler2 = no_events(handler3_mock)
        state = State()
        state._traverse_handlers(update='test_update',
                                 state_data={'test_data': 12345})
        handler1_mock.assert_not_called()
        handler2_mock.assert_not_called()
        handler3_mock.assert_called_once_with(
            state, 'test_update', Context(None, {'test_data': 12345}))

    def test_decorated_method_not_called(self):

        class State(BaseState):
            def handler1(self, update, data):
                pass

            def handler2(self, update, event_cls, data):
                pass

            def handler3(self, update, data):
                pass

        handler1_mock = MagicMock(__name__='handler1')
        handler2_mock = MagicMock(__name__='handler2')
        handler3_mock = MagicMock(__name__='handler2')

        State.handler1 = event(EventFalse1)(handler1_mock)
        State.handler2 = events(EventFalse2, EventTrue1)(handler2_mock)
        State.handler3 = no_events(handler3_mock)
        state = State()
        state._traverse_handlers(update='test_update',
                                 state_data={'test_data': 12345})
        handler1_mock.assert_not_called()
        handler2_mock.assert_called_once_with(state, 'test_update',
                                              Context(EventTrue1,
                                                      {'test_data': 12345}))
        handler3_mock.assert_not_called()


    def test_before_after_called(self):
        class State(BaseState):

            def handler(self, update, data):
                pass

            def before_handling(self, method_name, context):
                pass

            def after_handling(self, method_name, context):
                pass

        handler_mock = MagicMock(__name__='handler')
        after_mock = MagicMock()
        before_mock = MagicMock()
        State.handler = no_events(handler_mock)
        State.after_handling = after_mock
        State.before_handling = before_mock
        state = State()
        state._traverse_handlers(update='test_update',
                                 state_data={'test_data': 12345})
        before_mock.assert_called_once_with('handler', Context(None, {
            'test_data': 12345}))
        after_mock.assert_called_once_with('handler',
                                           Context(None, {'test_data': 12345}))

    def test_not_raises_EventHandled_exception(self):
        class State(BaseState):
            @no_events
            def handler(self, update, data):
                pass

        state = State()
        try:
            state.handler('update', 'data')
        except EventHandled:
            self.assertTrue(False)

    def test_has_attr(self):
        """Testing the presence of the `_event_handler` attribute on the decorated method."""

        class A:
            @decorators.no_events
            def foo(self, a, b, c):
                return a + b + c

            def foo1(self):
                return

        self.assertTrue(hasattr(A().foo, '_no_events'))
        self.assertFalse(hasattr(A().foo1, '_no_events'))




"""Модуль для работы с декораторами."""
import logging
from functools import wraps
from typing import Type

from pywood.context import Context
from pywood.events import BaseEvent
from pywood._exceptions import EventHandled

logger = logging.getLogger(__name__)


def event(event_cls: Type[BaseEvent]):
    """Декоратор для определения одного обрабатываемого события.

    :param event_cls: класс, определяющий событие, которое должен обработать \
    декорируемый метод
    :type event_cls: Type[BaseEvent]

    .. note:: blabla kjhkhjkhkjhkhkjhjkhjk
    """

    def decorator(method):
        method.__dict__['_event_handler'] = True

        @wraps(method)
        def wrapper(state, update, state_data):
            method_name = method.__name__
            if event_cls.happened(update):
                context = Context(event_cls, state_data)

                state.before_handling(method_name, context)
                method(state, update, context)
                state.after_handling(method_name, context)
                raise EventHandled

        return wrapper

    return decorator


def events(*event_classes):
    """Декоратор для определения нескольких обрабатываемых событий.

    :param event_classes: классы событий
    :type event_classes: классы событий в виде позиционных аргументов

    :param pass_event_cls: передавать ли в функцию обработчик событие. Функция\
     должна принимать данный аргумент.
    :type pass_event_cls: bool
    """

    def decorator(method):
        method.__dict__['_event_handler'] = True
        method_name = method.__name__

        @wraps(method)
        def wrapper(state, update, state_data):
            for event_cls in event_classes:
                if event_cls.happened(update):
                    context = Context(event_cls, state_data)

                    state.before_handling(method_name, context)
                    method(state, update, context)
                    state.after_handling(method_name, context)

                    raise EventHandled

        return wrapper

    return decorator


def no_events(method):
    """Декоратор для метода срабатывающего если не обрабатывались события."""
    method.__dict__['_no_events'] = True
    method_name = method.__name__

    @wraps(method)
    def wrapper(state, update, state_data):
        context = Context(None, state_data)
        state.before_handling(method_name, context)
        method(state, update, context)

        state.after_handling(method_name, context)

    return wrapper



"""Различные утилиты."""


def delete_nones(d: dict) -> dict:
    """Удалить рекурсивно элементы с None значениями.

    :param d: словарь в котором требуется рекурсивно удалить все элементы со
    значениями None
    :type d: dict
    """
    if not isinstance(d, (dict, list)):
        return d
    if isinstance(d, list):
        return [v for v in (delete_nones(v) for v in d) if v]
    return {k: v for k, v in ((k, delete_nones(v)) for k, v in d.items()) if v}


def full_qual_class_name(cls) -> str:
    """Получить полный путь класса.
    :param cls: класс, для которого необходимо получить полный путь
    :type cls: класс
    """
    return ".".join([cls.__module__, cls.__name__])



import unittest
from unittest.mock import Mock

from pywood._machine import Machine
from pywood.context import Context
from pywood.decorators import event, no_events, events
from pywood.events import BaseEvent
from pywood.states import BaseState


class EventTrue1(BaseEvent):
    @staticmethod
    def happened(update) -> bool:
        return True

class EventTrue2(BaseEvent):
    @staticmethod
    def happened(update) -> bool:
        return True

class EventTrue3(BaseEvent):
    @staticmethod
    def happened(update) -> bool:
        return True


class EventFalse1(BaseEvent):
    @staticmethod
    def happened(update) -> bool:
        return False

class EventFalse2(BaseEvent):
    @staticmethod
    def happened(update) -> bool:
        return False

class EventFalse3(BaseEvent):
    @staticmethod
    def happened(update) -> bool:
        return False



class TestMachine(unittest.TestCase):
    def test_method__get_state_from_state_name(self):
        class State1(BaseState):
            pass

        class State2(BaseState):
            pass

        class State3(BaseState):
            pass

        machine = Machine(current_state_getter=lambda update: State3,
                          states=[State1, State2, State3])

        self.assertIs(machine.get_state_from_state_name('State2'), State2)

    def test_method__get_current_state(self):
        class State1(BaseState):
            pass

        class State2(BaseState):
            pass

        class State3(BaseState):
            pass

        machine = Machine(current_state_getter=lambda update: State2,
                          states=[State1, State2, State3])
        self.assertIs(machine.get_current_state(state='State2')[0], State2)
        self.assertIs(machine.get_current_state(state=State2)[0], State2)
    def test_method_process(self):
        state1_method1_mock = Mock()
        state1_method2_mock = Mock()
        state1_method3_mock = Mock()
        state1_method4_mock = Mock()
        state1_method5_mock = Mock()
        state1_method6_mock = Mock()

        state2_method1_mock = Mock()
        state2_method2_mock = Mock()
        state2_method3_mock = Mock()
        state2_method4_mock = Mock()
        state2_method5_mock = Mock()
        state2_method6_mock = Mock()

        state3_method1_mock = Mock()
        state3_method2_mock = Mock()
        state3_method3_mock = Mock()
        state3_method4_mock = Mock()
        state3_method5_mock = Mock()
        state3_method6_mock = Mock()

        state4_method1_mock = Mock()
        state4_method2_mock = Mock()
        state4_method3_mock = Mock()
        state4_method4_mock = Mock()
        state4_method5_mock = Mock()
        state4_method6_mock = Mock()

        before_handling_mock_state1 = Mock()
        before_handling_mock_state2 = Mock()
        before_handling_mock_state3 = Mock()
        before_handling_mock_state4 = Mock()
        after_handling_mock_state1 = Mock()
        after_handling_mock_state2 = Mock()
        after_handling_mock_state3 = Mock()
        after_handling_mock_state4 = Mock()

        class Event1(BaseEvent):
            @staticmethod
            def happened(update) -> bool:
                if update == 'event1':
                    return True
                else:
                    return False

        class Event2(BaseEvent):
            @staticmethod
            def happened(update) -> bool:
                if update == 'event2':
                    return True
                else:
                    return False

        class Event3(BaseEvent):
            @staticmethod
            def happened(update) -> bool:
                if update == 'event3':
                    return True
                else:
                    return False

        class Event4(BaseEvent):
            @staticmethod
            def happened(update) -> bool:
                if update == 'event4':
                    return True
                else:
                    return False

        class Event5(BaseEvent):
            @staticmethod
            def happened(update) -> bool:
                if update == 'event5':
                    return True
                else:
                    return False

        class Event6(BaseEvent):
            @staticmethod
            def happened(update=None) -> bool:
                if update == 'event6':
                    return True
                else:
                    return False

        class TestState1(BaseState):
            @event(Event4)
            def method1(self, update, context):
                state1_method1_mock()

            @no_events
            def method2(self, update, context):
                state1_method2_mock()

            def method3(self, update, context):
                state1_method3_mock()

            @events(Event1, Event2, Event3)
            def method4(self, update, context):
                state1_method4_mock()

            @events(Event5)
            def method5(self, update, context):
                state1_method5_mock()

            @events(Event5)
            def method6(self, update, context):
                state1_method6_mock()

            def after_handling(self, method_name, context):
                after_handling_mock_state1(method_name, context)

            def before_handling(self, method_name, context):
                before_handling_mock_state1(method_name, context)

        class TestState2(BaseState):
            @event(Event5)
            def method1(self, update, context):
                state2_method1_mock()

            @no_events
            def method2(self, update, context):
                state2_method2_mock()

            def method3(self, update, context):
                state2_method3_mock()

            @events(Event2, Event3)
            def method4(self, update, context):
                state2_method4_mock()

            @events(Event1)
            def method5(self, update, context):
                state2_method5_mock()

            def method6(self):
                state2_method6_mock()

            def after_handling(self, method_name, context):
                after_handling_mock_state2(method_name, context)

            def before_handling(self, method_name, context):
                before_handling_mock_state2(method_name, context)

        class TestState3(BaseState):
            @events(Event4, Event5, Event6)
            def method1(self, update, context):
                state3_method1_mock()

            @no_events
            def method2(self, update, context):
                state3_method2_mock()

            def method3(self, update, context):
                state3_method3_mock()

            @events(Event1, Event2, Event3)
            def method4(self, update, context):
                state3_method4_mock()

            def method5(self, update, context):
                state3_method5_mock()

            def method6(self, update, context):
                state3_method6_mock()

            def after_handling(self, method_name, context):
                after_handling_mock_state3(method_name, context)

            def before_handling(self, method_name, context):
                before_handling_mock_state3(method_name, context)

        class TestState4(BaseState):
            @event(Event5)
            def method1(self, update, context):
                state4_method1_mock()

            def method2(self, update, context):
                state4_method2_mock()

            def method3(self, update, context):
                state4_method3_mock()

            @events(Event2, Event3)
            def method4(self, update, context):
                state4_method4_mock()

            @events(Event1)
            def method5(self, update, context):
                state4_method5_mock()

            def method6(self, update, context):
                state4_method6_mock()

            def after_handling(self, method_name, context):
                after_handling_mock_state4(method_name, context)

            def before_handling(self, method_name, context):
                before_handling_mock_state4(method_name, context)

        machine = Machine(current_state_getter=lambda update: TestState1,
                          states=[TestState1, TestState2, TestState3])
        machine.process(update='event5',
                        state_attrs={'bla': 123, 'tiktak': 'Barack Trump'})
        state1_method1_mock.assert_not_called()
        state1_method2_mock.assert_not_called()
        state1_method3_mock.assert_not_called()
        state1_method4_mock.assert_not_called()
        state1_method5_mock.assert_called_once()
        state1_method6_mock.assert_not_called()
        before_handling_mock_state1.assert_called_once_with('method5',
                                                            Context(Event5,
                                                                    None))
        after_handling_mock_state1.assert_called_once_with('method5',
                                                           Context(Event5,
                                                                   None))

        machine = Machine(current_state_getter=lambda update: TestState2,
                          states=[TestState1, TestState2, TestState3])
        machine.process(update='event6')
        state2_method1_mock.assert_not_called()
        state2_method2_mock.assert_called_once()
        state2_method3_mock.assert_not_called()
        state2_method4_mock.assert_not_called()
        state2_method5_mock.assert_not_called()
        state2_method6_mock.assert_not_called()
        before_handling_mock_state2.assert_called_once_with('method2',
                                                            Context(None,
                                                                    None))
        after_handling_mock_state2.assert_called_once_with('method2',
                                                           Context(None, None))

        machine = Machine(current_state_getter=lambda update: TestState3,
                          states=[TestState1, TestState2, TestState3])
        machine.process(update='event5')
        state3_method1_mock.assert_called_once()
        state3_method2_mock.assert_not_called()
        state3_method3_mock.assert_not_called()
        state3_method4_mock.assert_not_called()
        state3_method5_mock.assert_not_called()
        state3_method6_mock.assert_not_called()
        before_handling_mock_state3.assert_called_once_with('method1',
                                                            Context(Event5,
                                                                    None))
        after_handling_mock_state3.assert_called_once_with('method1',
                                                           Context(Event5,
                                                                   None))

        machine = Machine(current_state_getter=lambda update: TestState4,
                          states=[TestState1, TestState2, TestState3])
        machine.process(update='event6')
        state4_method1_mock.assert_not_called()
        state4_method2_mock.assert_not_called()
        state4_method3_mock.assert_not_called()
        state4_method4_mock.assert_not_called()
        state4_method5_mock.assert_not_called()
        state4_method6_mock.assert_not_called()
        before_handling_mock_state4.assert_not_called()
        after_handling_mock_state4.assert_not_called()

    def test_state_has_passed_from_handle_update_attrs(self):
        class State1(BaseState):
            pass

        class State2(BaseState):
            pass

        class State3(BaseState):
            def _traverse_handlers(self2, *args, **kwargs):
                self.assertTrue(self2.db_conn == 'db_connection')
                self.assertTrue(self2.bot_api == 'bot_api')

        class State4(BaseState):
            pass

        machine = Machine(current_state_getter=lambda update: State3,
                          states=[State1, State2, State3, State4])
        machine.process(update=Mock(),
                        state_attrs={
                            'db_conn': 'db_connection',
                            'bot_api': 'bot_api',
                        }
                        )


from typing import List


def is_command(update):
    try:
        if not update.message.text.startswith('/'):
            return False
    except AttributeError:
        return False

    try:
        entities = update.message.entities
    except AttributeError:
        return False

    def check_entities(entities: List):
        def entity_is_command(entity) -> bool:
            if entity.type == 'bot_command':
                return True
            else:
                return False

        if entities:
            if len(entities) != 1:
                return False
            else:
                return entity_is_command(entities[0])
        else:
            return False

    return check_entities(entities)

def is_start_command(update):
    if not is_command(update):
        return False
    else:
        if update.message.text == '/start':
            return True
        else:
            return False



import unittest
from unittest.mock import MagicMock, patch

from pywood.context import Context
from pywood.decorators import event, events, no_events
from pywood.events import BaseEvent
from pywood.states import BaseState


class Event1(BaseEvent):
    @staticmethod
    def happened(update=None) -> bool:
        return True


class Event2(BaseEvent):
    @staticmethod
    def happened(update=None) -> bool:
        return False


class Event3(BaseEvent):
    @staticmethod
    def happened(update=None) -> bool:
        return True


class Event4(BaseEvent):
    @staticmethod
    def happened(update=None) -> bool:
        return False


class Event5(BaseEvent):
    @staticmethod
    def happened(update=None) -> bool:
        return False


class Event6(BaseEvent):
    @staticmethod
    def happened(update=None) -> bool:
        return True


class TestState1(BaseState):
    @event(Event4)
    def method1(self, update, context):
        raise NotImplementedError

    @no_events
    def method2(self, update, context):
        raise NotImplementedError

    def method3(self, update, context):
        raise NotImplementedError

    @events(Event1, Event2, Event3)
    def method4(self, update, context):
        raise NotImplementedError

    @events(Event5)
    def method5(self, update, context):
        raise NotImplementedError

    @events(Event5)
    def method6(self, update, context):
        raise NotImplementedError


class TestStateMethods(unittest.TestCase):
    def test_method__handlers(self):
        state = TestState1()
        self.assertListEqual([method[0] for method in state._handlers()],
                             ['method1', 'method4', 'method5', 'method6']
                             )

    def test_method__no_event_handler(self):
        state = TestState1()
        self.assertEqual(state._no_event_handler()[0], 'method2')

    @patch.object(Event1, 'happened', return_value=False)
    @patch.object(Event2, 'happened', return_value=False)
    @patch.object(Event3, 'happened', return_value=False)
    @patch.object(Event4, 'happened', return_value=False)
    @patch.object(Event5, 'happened', return_value=False)
    @patch.object(Event6, 'happened', return_value=False)
    def test__traverse_handlers_called_no_events_handler(self, *args):
        method2_mock = MagicMock(__name__='method2')

        with patch.object(TestState1, 'method2', no_events(method2_mock)):
            state = TestState1()
            state._traverse_handlers('update', 'state_data')
            method2_mock.assert_called_once_with(state, 'update',
                                                 Context(None, 'state_data'))

    def test__traverse_handlers_called_first_handler(self):
        method4_mock = MagicMock(__name__='method4')

        with patch.object(TestState1, 'method4',
                          events(Event1, Event2, Event3)(method4_mock)):
            state = TestState1()
            state._traverse_handlers('update', 'state_data')
            method4_mock.assert_called_once_with(state, 'update',
                                                 Context(Event1, 'state_data'))

    def test_raises(self):
        class State(BaseState):
            def handler(self, update, context):
                pass

        state = State()
        state._traverse_handlers('update', 'state_data')

    def test_state_name(self):
        class MyCustomState(BaseState):
            pass

        state = MyCustomState()
        self.assertEqual(state.state_name, 'MyCustomState')



import unittest
from unittest.mock import Mock

from pywood import BaseState
from pywood.context import Context
from pywood.decorators import event, events, no_events
from pywood.events import BaseEvent


class EventTrue1(BaseEvent):
    @staticmethod
    def happened(update) -> bool:
        return True

class EventTrue2(BaseEvent):
    @staticmethod
    def happened(update) -> bool:
        return True

class EventTrue3(BaseEvent):
    @staticmethod
    def happened(update) -> bool:
        return True


class EventFalse1(BaseEvent):
    @staticmethod
    def happened(update) -> bool:
        return False

class EventFalse2(BaseEvent):
    @staticmethod
    def happened(update) -> bool:
        return False

class EventFalse3(BaseEvent):
    @staticmethod
    def happened(update) -> bool:
        return False

class EventFalse4(BaseEvent):
    @staticmethod
    def happened(update) -> bool:
        return False

class TestEventDecorator(unittest.TestCase):
    def test_no_events_called(self):
        mock1 = Mock()
        mock2 = Mock()
        mock3 = Mock()
        mock4 = Mock()

        class State(BaseState):
            @event(EventFalse1)
            def handle1(self, update, context):
                mock1(self, update, context)

            @event(EventFalse2)
            def handle2(self, update, context):
                mock2(self, update, context)

            @events(EventFalse3, EventFalse4)
            def handle3(self, update, context):
                mock3(self, update, context)

            @no_events
            def handle4(self, update, context):
                mock4(self, update, context)

        state = State()
        state._traverse_handlers('update', 'data')
        mock1.assert_not_called()
        mock2.assert_not_called()
        mock3.assert_not_called()
        mock4.assert_called_once_with(state, 'update', Context(None, 'data'))

    def test_events_called(self):
        mock1 = Mock()
        mock2 = Mock()
        mock3 = Mock()
        mock4 = Mock()

        class State(BaseState):
            @event(EventFalse1)
            def handle1(self, update, context):
                mock1(self, update, context)

            @event(EventFalse2)
            def handle2(self, update, context):
                mock2(self, update, context)

            @events(EventFalse3, EventTrue1)
            def handle3(self, update, context):
                mock3(self, update, context)

            @no_events
            def handle4(self, update, context):
                mock4(self, update, context)

        state = State()
        state._traverse_handlers('update', 'data')
        mock1.assert_not_called()
        mock2.assert_not_called()
        mock3.assert_called_once_with(state, 'update',
                                      Context(EventTrue1, 'data'))
        mock4.assert_not_called()


    def test_event_called(self):
        mock1 = Mock()
        mock2 = Mock()
        mock3 = Mock()
        mock4 = Mock()

        class State(BaseState):
            @event(EventFalse1)
            def handle1(self, update, context):
                mock1(self, update, context)

            @event(EventTrue1)
            def handle2(self, update, context):
                mock2(self, update, context)

            @events(EventFalse3, EventTrue1)
            def handle3(self, update, context):
                mock3(self, update, context)

            @no_events
            def handle4(self, update, data):
                mock4(self, update, data)

        state = State()
        state._traverse_handlers('update', state_data='data')
        mock1.assert_not_called()
        mock2.assert_called_once_with(state, 'update',
                                      Context(EventTrue1, 'data'))
        mock3.assert_not_called()
        mock4.assert_not_called()




from codecs import open

import setuptools
from distutils.core import setup

__title__ = 'pywood'
__description__ = 'Lightweight framework to build Telegram bots.'
__url__ = 'https://github.com/pynista/pywood'
__version__ = '0.0.6'
__author__ = 'Dzmitry Maliuzhenets'
__author_email__ = 'dzmitrymaliuzhenets@gmail.com'
__license__ = 'MIT'
__copyright__ = 'Copyright 2020 Dzmitry Maliuzhenets'

with open('README.md', 'r', 'utf-8') as f:
    readme = f.read()


requirements = [
    'prettyprinter',
    'wheel',
    'telegrambotapiwrapper'
]

setup(
    python_requires='~=3.7',
    name=__title__,
    version=__version__,
    author=__author__,
    author_email=__author_email__,
    description=__description__,
    long_description=readme,
    long_description_content_type='text/markdown',
    url=__url__,
    install_requires=requirements,
    packages=setuptools.find_packages(),
    setup_requires=['wheel'],
    keywords='telegram api bot framework durov',
    classifiers=[

        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Natural Language :: Russian',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Operating System :: OS Independent',
    ],
)



"""Модуль для работы с состояниями."""
import inspect
from abc import ABCMeta
from typing import List, Callable, Tuple

from pywood._exceptions import EventHandled


class BaseState(metaclass=ABCMeta):
    """The base class for all states.

    Create custom states by inheriting from it.
    """

    @property
    def state_name(self):
        return self.__class__.__name__

    def _handlers(self) -> List[Tuple[str, Callable]]:
        """Получить методы которые декорированы декораторами event и events."""

        def predicate(value) -> bool:
            return inspect.ismethod(value) and hasattr(value, '_event_handler')

        return inspect.getmembers(self, predicate=predicate)

    def _no_event_handler(self) -> Tuple[str, Callable]:
        """Получить метод, который декорирован декоратором no_events"""

        def predicate(value):
            return inspect.ismethod(value) and hasattr(value, '_no_events')

        return inspect.getmembers(self, predicate=predicate)[0]

    def after_handling(self, method_name, context):
        """Метод, который выполняется после обработки события.

        :param method_name: имя метода
        :type method_name: str

        :param event_cls: класс события
        :type event_cls: класс
        """
        pass

    def before_handling(self, method_name, context):
        """Метод, который выполняется перед обработкой события.

        :param method_name: имя метода
        :type method_name: str

        :param event_cls: класс события
        :type event_cls: класс
        """
        pass

    def _traverse_handlers(self, update, state_data):
        """Пройти через обработчики."""

        for method_name, method in self._handlers():
            try:
                method(update, state_data)
            except EventHandled:
                return
        else:
            try:
                method_name, method = self._no_event_handler()
            except IndexError:
                pass
            else:
                method(update, state_data)



"""Модуль для работы с pooling."""
from typing import Callable

from requests import Timeout
from telegrambotapiwrapper import Api




def _process_existing_updates(bot_api, handler):
    offset = None

    while True:
        existing_updates = bot_api.get_updates(offset=offset)
        if existing_updates:
            for update in existing_updates:
                handler(update)
                offset = update.update_id + 1
        else:
            return


def listen(token: str,
           handler: Callable,
           process_existing: bool = False,
           ):
    """

    :param token: токен
    :type token: str

    :param handler: функция-обработчик
    :type handler: callable, принимающее один аргумент вида `update <https://core.telegram.org/bots/api#update>`_

    :param process_existing: (optional) обработать уже скопившиеся обновления
        или нет
    :type process_existing: bool

    .. Note:: Если process_existing равно False, то хотя скопившиеся
        обновления не будут обработаны обработчиком handler, но они будут
        `подтверждены <https://core.telegram.org/bots/faq#long-polling-gives-me-the-same-updates-again-and-again>`_
    """
    timeout = 20
    bot_api = Api(token=token)
    if process_existing:
        _process_existing_updates(bot_api, handler)
    else:
        _process_existing_updates(bot_api, lambda update: None)

    offset = None
    while True:
        try:
            updates = bot_api.get_updates(offset=offset, timeout=timeout)
        except Timeout:
            continue
        for update in updates:
            handler(update)
            offset = update.update_id + 1



"""Модуль для работы с событиями."""
from abc import ABCMeta, abstractmethod


class BaseEvent(metaclass=ABCMeta):
    """The base class for all events.

    Create custom events by inheriting from it.
    """

    @staticmethod
    @abstractmethod
    def happened(update) -> bool:
        """Проверить, сработало ли событие для данного `update`."""
        raise NotImplementedError

    @property
    def name(self) -> str:
        """Получить имя класса события.

        :returns: имя класса события, например `IncomingText`
        :rtype: str
        """
        return self.__class__.__name__

    def __repr__(self):
        """Получить представление класса события как имя класса.

        :returns: имя класса события со скобками, например `IncomingText()`
        :rtype: str
        """
        return f'{self.__class__.__name__}()'



from copy import deepcopy


def get_anonymized_update(update):
    new_update = deepcopy(update)
    new_update.update_id = 123456789
    new_update.message.chat.id = 123456789
    new_update.message.chat.username = 'Username'
    new_update.message.chat.first_name = 'FirstName'
    new_update.message.from_.id = 123456789
    new_update.message.from_.username = 'blablabla'
    new_update.message.from_.first_name = 'hurraaaa'
    return new_update



import datetime
from dataclasses import asdict

from prettyprinter import prettyprinter

from pywood.pooling import listen
from pywood.utils import delete_nones


def print_incoming_updates(token: str,
                           print_existing: bool = False,
                           print_with_nones: bool = True):
    """Слушать и печатать входящие обновления.

    :param token: токен
    :type token: str

    :param print_existing: (optional) распечатать уже скопившиеся updates
    :type print_existing: bool

    :param delay: (optional) задержка в секундах между обращениями к Telegram Bot Api
    :type delay: int

    :param print_with_nones: (optional) включать поля, которые отсутствут в полученных объектах
    :type print_with_nones: bool
    """

    def w_nones(update):
        now = datetime.datetime.now()
        print('---------{}---------'.format(now.strftime("%Y-%m-%d %H:%M:%S")))
        prettyprinter.pprint(asdict(update))

    def wo_nones(update):
        now = datetime.datetime.now()
        print('---------{}---------'.format(now.strftime("%Y-%m-%d %H:%M:%S")))
        prettyprinter.pprint(delete_nones(asdict(update)))

    handler = w_nones if print_with_nones else wo_nones
    listen(token, handler, print_existing)
